use crate::prelude::*;
use anyhow::Result;
use std::collections::HashMap;

#[derive(Debug)]
pub enum EngineEvent<'a> {
    MouseMotion(u32, u32),
    MouseButtonDown(MouseButton),
    MouseButtonUp(MouseButton),
    MouseWheel(i32, i32, MouseWheelDirection),

    KeyDown(Keycode),
    KeyUp(Keycode),

    ButtonDown(ControllerButton),
    ButtonUp(ControllerButton),
    ControllerAxis(ControllerAxis),

    ControllerAdded(&'a Controller),
    ControllerRemoved(&'a Controller),

    FileDrop(String),
}

pub trait EventConsumer: Send + Sync + 'static {
    fn event(&mut self, world: &mut World, event: EngineEvent<'_>) -> Result<()>;
}

pub struct InputMap {
    pub direction_mapping: HashMap<Keycode, GuiDirection>,
}

impl Engine {
    pub(crate) fn event<T: EventConsumer>(
        world: &mut World,
        context: &Context,
        gui_handler: &GuiHandler,
        input: &InputMap,
        consumer: &mut T,
        event: Event<'_>,
    ) -> anyhow::Result<()> {
        match event {
            Event::MouseMotion(x, y) => {
                gui_handler.set_mouse_pos(x, y)?;
                consumer.event(world, EngineEvent::MouseMotion(x, y))?;
            }
            Event::MouseButtonDown(mouse_button) => {
                if !gui_handler.mouse_down(mouse_button)? {
                    consumer.event(world, EngineEvent::MouseButtonDown(mouse_button))?;
                }
            }
            Event::MouseButtonUp(mouse_button) => {
                if !gui_handler.mouse_up(mouse_button)? {
                    consumer.event(world, EngineEvent::MouseButtonUp(mouse_button))?;
                }
            }
            Event::MouseWheel(x, y, direction) => {
                consumer.event(world, EngineEvent::MouseWheel(x, y, direction))?
            }
            Event::KeyDown(keycode) => {
                Self::key_down_event(world, context, gui_handler, input, consumer, keycode)?;
            }
            Event::KeyUp(keycode) => {
                Self::key_up_event(world, gui_handler, input, consumer, keycode)?;
            }
            Event::TextInput(text) => {
                Self::text_input(gui_handler, text)?;
            }
            Event::ControllerButtonDown(_controller, button) => {
                Self::button_down_event(world, gui_handler, consumer, button)?;
            }
            Event::ControllerButtonUp(_controller, button) => {
                Self::button_up_event(world, consumer, button)?;
            }
            Event::ControllerAxis(controller) => {
                if !gui_handler.check_navigatable()? {
                    Self::axis_event(world, consumer, &controller)?
                } else {
                    gui_handler.update_selection(controller.direction())?;
                }
            }
            Event::ControllerAdded(controller) => {
                Self::controller_added(world, consumer, controller)?
            }
            Event::ControllerRemoved(controller) => {
                Self::controller_removed(world, consumer, controller)?
            }

            Event::JoystickAxis(_joystick) => todo!(),
            Event::JoystickButtonDown(_joystick) => todo!(),
            Event::JoystickButtonUp(_joystick) => todo!(),
            Event::JoystickAdded(_joystick) => todo!(),
            Event::JoystickRemoved(_joystick) => todo!(),

            Event::FileDrop(filename) => consumer.event(world, EngineEvent::FileDrop(filename))?,
        }

        Ok(())
    }
}

impl Engine {
    #[inline]
    fn controller_added<T: EventConsumer>(
        world: &mut World,
        consumer: &mut T,
        controller: &Controller,
    ) -> Result<()> {
        consumer.event(world, EngineEvent::ControllerAdded(controller))?;

        Ok(())
    }

    #[inline]
    fn controller_removed<T: EventConsumer>(
        world: &mut World,
        consumer: &mut T,
        controller: &Controller,
    ) -> Result<()> {
        consumer.event(world, EngineEvent::ControllerRemoved(controller))?;

        Ok(())
    }

    #[inline]
    fn key_up_event<T: EventConsumer>(
        world: &mut World,
        gui_handler: &GuiHandler,
        input: &InputMap,
        consumer: &mut T,
        keycode: Keycode,
    ) -> Result<()> {
        if input.direction_mapping.get(&keycode).is_some()
            && gui_handler.update_selection(GuiDirection::None)?
        {
            return Ok(());
        }

        consumer.event(world, EngineEvent::KeyUp(keycode))?;

        Ok(())
    }

    #[inline]
    fn key_down_event<T: EventConsumer>(
        world: &mut World,
        context: &Context,
        gui_handler: &GuiHandler,
        input: &InputMap,
        consumer: &mut T,
        keycode: Keycode,
    ) -> Result<()> {
        if let Some(direction) = input.direction_mapping.get(&keycode) {
            if gui_handler.update_selection(*direction)? {
                return Ok(());
            }
        }

        match keycode {
            Keycode::Backspace => {
                if gui_handler.remove_char()? {
                    return Ok(());
                }
            }
            Keycode::Return => {
                if gui_handler.accept_selection()? {
                    return Ok(());
                }
            }
            Keycode::Escape => {
                if gui_handler.decline_topgui()? {
                    return Ok(());
                }
            }
            Keycode::V => {
                if let Some(writeable) = gui_handler.writeable()? {
                    if let Some(content) = context.window_config().clipboard_content()? {
                        writeable.set_text(content)?;
                    }
                }
            }

            _ => (),
        }

        consumer.event(world, EngineEvent::KeyDown(keycode))?;

        Ok(())
    }

    #[inline]
    fn button_up_event<T: EventConsumer>(
        world: &mut World,
        consumer: &mut T,
        button: ControllerButton,
    ) -> Result<()> {
        consumer.event(world, EngineEvent::ButtonUp(button))?;

        Ok(())
    }

    #[inline]
    fn button_down_event<T: EventConsumer>(
        world: &mut World,
        gui_handler: &GuiHandler,
        consumer: &mut T,
        button: ControllerButton,
    ) -> Result<()> {
        if gui_handler.check_navigatable()? {
            Self::check_button_down(world, gui_handler, consumer, button)?;
        } else {
            consumer.event(world, EngineEvent::ButtonDown(button))?;
        }

        Ok(())
    }

    #[inline]
    fn axis_event<T: EventConsumer>(
        world: &mut World,
        consumer: &mut T,
        controller: &Controller,
    ) -> Result<()> {
        consumer.event(
            world,
            EngineEvent::ControllerAxis(controller.controller_axis()),
        )?;

        Ok(())
    }

    #[inline]
    fn check_button_down<T: EventConsumer>(
        world: &mut World,
        gui_handler: &GuiHandler,
        consumer: &mut T,
        button: ControllerButton,
    ) -> Result<()> {
        match button {
            ControllerButton::A => {
                if gui_handler.accept_selection()? {
                    return Ok(());
                }
            }

            ControllerButton::B => {
                if gui_handler.decline_topgui()? {
                    return Ok(());
                }
            }

            ControllerButton::RightButton => {
                if gui_handler.next_tab_topgui(false)? {
                    return Ok(());
                }
            }

            ControllerButton::LeftButton => {
                if gui_handler.previous_tab_topgui(false)? {
                    return Ok(());
                }
            }

            ControllerButton::RightTrigger => {
                if gui_handler.next_tab_topgui(true)? {
                    return Ok(());
                }
            }

            ControllerButton::LeftTrigger => {
                if gui_handler.previous_tab_topgui(true)? {
                    return Ok(());
                }
            }

            ControllerButton::DPadDown => {
                let selection_res = gui_handler.update_selection(GuiDirection::Down)?;
                gui_handler.update_selection(GuiDirection::None)?;

                if selection_res {
                    return Ok(());
                }
            }

            ControllerButton::DPadUp => {
                let selection_res = gui_handler.update_selection(GuiDirection::Up)?;
                gui_handler.update_selection(GuiDirection::None)?;

                if selection_res {
                    return Ok(());
                }
            }

            ControllerButton::DPadRight => {
                let selection_res = gui_handler.update_selection(GuiDirection::Right)?;
                gui_handler.update_selection(GuiDirection::None)?;

                if selection_res {
                    return Ok(());
                }
            }

            ControllerButton::DPadLeft => {
                let selection_res = gui_handler.update_selection(GuiDirection::Left)?;
                gui_handler.update_selection(GuiDirection::None)?;

                if selection_res {
                    return Ok(());
                }
            }

            _ => (),
        }

        if !gui_handler.accept_custom_selection(button)? {
            consumer.event(world, EngineEvent::ButtonDown(button))?;
        }

        Ok(())
    }

    #[inline]
    fn text_input(gui_handler: &GuiHandler, text: String) -> Result<()> {
        if let Some(writeable) = gui_handler.writeable()? {
            for c in text.chars() {
                writeable.add_letter(c)?;
            }
        }

        Ok(())
    }
}
