use crate::prelude::*;

use anyhow::Result;
use utilities::{impl_reprc, prelude::cgmath::Vector3};

use super::{
    super::{
        ExtensionCheck, RaytracerDesciptorLayouts, RenderingFrontEnd,
        shared::position_buffer_reader::PositionBuffer,
    },
    RenderingCore,
    animator::ComputeBasedAnimator,
    light::{LightData, RayTracerLight},
};

use std::sync::{Arc, Mutex};

const STARTING_DESCRIPTOR_COUNT: u32 = 64;
const Z_NEAR: f32 = 0.01;
const Z_FAR: f32 = 100.0;

impl_reprc!(
    #[derive(Copy)]
    pub struct LightInfo {
        #[assume_reprc]
        light_count: u32,
    }
);

pub struct RayTracer<T>
where
    T: RenderingCore,
{
    render_scale: f32,

    device: Arc<Device>,
    queue: Arc<Mutex<Queue>>,

    view: View,

    // target for ray tracing, is going to be blit into swapchain image
    target_image: Arc<Image>,
    position_image: Arc<Image>,

    position_buffer: PositionBuffer,

    // scene object data (objects, materials, textures)
    scene_desc_set: Arc<DescriptorSet>,
    max_material_count: u32,
    material_desc_set: Arc<DescriptorSet>,
    max_texture_count: u32,
    texture_desc_set: Arc<DescriptorSet>,

    // lights
    lights: Vec<RayTracerLight>,
    light_descriptor_set: Arc<DescriptorSet>,

    light_info_buffer: Arc<Buffer<LightInfo>>,
    light_info_desc_set: Arc<DescriptorSet>,

    // handle to iterate over scene data, executes animation
    animator: ComputeBasedAnimator,

    // screen parameter
    width: u32,
    height: u32,

    renderer: T,
}

impl<T> RayTracer<T>
where
    T: RenderingCore,
{
    pub fn new(
        device: &Arc<Device>,
        queue: &Arc<Mutex<Queue>>,
        (screen_width, screen_height): (f32, f32),
        image_count: usize,
        graphics_info: GraphicsInfo,
        raytrace_info: RaytracingInfo,
    ) -> Result<RayTracer<T>> {
        let render_scale = graphics_info.render_scale;

        let width = (screen_width * render_scale) as u32;
        let height = (screen_height * render_scale) as u32;

        let target_image = Self::create_target_image(device, queue, width, height)?;
        let position_image = Self::create_position_image(device, queue, width, height)?;

        let proj = perspective(45.0, screen_width / screen_height, Z_NEAR, Z_FAR);
        let view = View::new(device, proj, 45.0, false)?;

        let scene_desc_set = Self::create_scene_descriptor_set(device)?;
        let material_desc_set =
            Self::create_material_descriptor_set(device, STARTING_DESCRIPTOR_COUNT)?;
        let texture_desc_set =
            Self::create_texture_descriptor_set(device, STARTING_DESCRIPTOR_COUNT)?;

        let light_descriptor_set = Self::create_light_descriptor(device, raytrace_info.max_lights)?;
        let (light_buffer, light_desc_set) = Self::create_light_info(device)?;

        let animator = ComputeBasedAnimator::new(device, queue, image_count)?;

        Ok(RayTracer {
            render_scale: graphics_info.render_scale,

            renderer: T::new(
                device,
                queue,
                raytrace_info,
                &view,
                &target_image,
                &position_image,
                vec![
                    &scene_desc_set,
                    &material_desc_set,
                    &texture_desc_set,
                    &light_descriptor_set,
                    &light_desc_set,
                ],
            )?,

            device: device.clone(),
            queue: queue.clone(),

            view,

            position_buffer: PositionBuffer::new(&position_image),

            target_image,
            position_image,

            scene_desc_set,
            max_material_count: STARTING_DESCRIPTOR_COUNT,
            material_desc_set,
            max_texture_count: STARTING_DESCRIPTOR_COUNT,
            texture_desc_set,

            lights: Vec::new(),
            light_descriptor_set,

            light_info_buffer: light_buffer,
            light_info_desc_set: light_desc_set,

            animator,

            width,
            height,
        })
    }

    #[inline]
    fn create_target_image(
        device: &Arc<Device>,
        queue: &Arc<Mutex<Queue>>,
        width: u32,
        height: u32,
    ) -> Result<Arc<Image>> {
        let target_image = Image::empty(
            width,
            height,
            VK_IMAGE_USAGE_TRANSFER_SRC_BIT | VK_IMAGE_USAGE_STORAGE_BIT,
            VK_SAMPLE_COUNT_1_BIT,
        )
        .format(VK_FORMAT_R8G8B8A8_UNORM)
        .build(device, queue)?;

        Image::convert_layout(&target_image, VK_IMAGE_LAYOUT_GENERAL)?;

        Ok(target_image)
    }

    #[inline]
    fn create_position_image(
        device: &Arc<Device>,
        queue: &Arc<Mutex<Queue>>,
        width: u32,
        height: u32,
    ) -> Result<Arc<Image>> {
        let depth_buffer = Image::empty(
            width,
            height,
            VK_IMAGE_USAGE_STORAGE_BIT,
            VK_SAMPLE_COUNT_1_BIT,
        )
        .format(VK_FORMAT_R32G32B32A32_SFLOAT)
        .add_usage(VK_IMAGE_USAGE_TRANSFER_SRC_BIT)
        .build(device, queue)?;

        Image::convert_layout(&depth_buffer, VK_IMAGE_LAYOUT_GENERAL)?;

        Ok(depth_buffer)
    }

    fn create_scene_descriptor_set(device: &Arc<Device>) -> Result<Arc<DescriptorSet>> {
        let desc_layout = RaytracerDesciptorLayouts::vertex_type_descriptor_layout(device.clone())?;

        let descriptor_pool = DescriptorPool::builder()
            .set_layout(desc_layout)
            .set_flags(VK_DESCRIPTOR_POOL_CREATE_UPDATE_AFTER_BIND_BIT_EXT)
            .build(device.clone())?;

        DescriptorPool::prepare_set(&descriptor_pool).allocate()
    }

    fn create_material_descriptor_set(
        device: &Arc<Device>,
        max_size: u32,
    ) -> Result<Arc<DescriptorSet>> {
        let desc_layout = DescriptorSetLayout::builder()
            // materials
            .add_layout_binding(
                0,
                VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER,
                VK_SHADER_STAGE_CLOSEST_HIT_BIT_KHR
                    | VK_SHADER_STAGE_COMPUTE_BIT
                    | VK_SHADER_STAGE_ANY_HIT_BIT_KHR,
                VK_DESCRIPTOR_BINDING_VARIABLE_DESCRIPTOR_COUNT_BIT_EXT
                    | VK_DESCRIPTOR_BINDING_PARTIALLY_BOUND_BIT_EXT,
            )
            .change_descriptor_count(max_size)
            .build(device.clone())?;

        let descriptor_pool = DescriptorPool::builder()
            .set_layout(desc_layout)
            .set_flags(VK_DESCRIPTOR_POOL_CREATE_UPDATE_AFTER_BIND_BIT_EXT)
            .build(device.clone())?;

        DescriptorPool::prepare_set(&descriptor_pool)
            .set_variable_descriptor_counts(&[max_size])
            .allocate()
    }

    fn create_texture_descriptor_set(
        device: &Arc<Device>,
        max_size: u32,
    ) -> Result<Arc<DescriptorSet>> {
        let desc_layout = DescriptorSetLayout::builder()
            // textures
            .add_layout_binding(
                0,
                VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER,
                VK_SHADER_STAGE_CLOSEST_HIT_BIT_KHR
                    | VK_SHADER_STAGE_COMPUTE_BIT
                    | VK_SHADER_STAGE_ANY_HIT_BIT_KHR,
                VK_DESCRIPTOR_BINDING_VARIABLE_DESCRIPTOR_COUNT_BIT_EXT
                    | VK_DESCRIPTOR_BINDING_PARTIALLY_BOUND_BIT_EXT,
            )
            .change_descriptor_count(max_size)
            .build(device.clone())?;

        let descriptor_pool = DescriptorPool::builder()
            .set_layout(desc_layout)
            .build(device.clone())?;

        DescriptorPool::prepare_set(&descriptor_pool)
            .set_variable_descriptor_counts(&[max_size])
            .allocate()
    }

    fn create_light_descriptor(
        device: &Arc<Device>,
        max_lights: u32,
    ) -> Result<Arc<DescriptorSet>> {
        let layout =
            RaytracerDesciptorLayouts::light_descriptor_layout(device.clone(), max_lights)?;

        let descriptor_pool = DescriptorPool::builder()
            .set_layout(layout)
            .set_flags(VK_DESCRIPTOR_POOL_CREATE_UPDATE_AFTER_BIND_BIT_EXT)
            .build(device.clone())?;

        DescriptorPool::prepare_set(&descriptor_pool)
            .set_variable_descriptor_counts(&[max_lights])
            .allocate()
    }

    fn create_light_info(
        device: &Arc<Device>,
    ) -> Result<(Arc<Buffer<LightInfo>>, Arc<DescriptorSet>)> {
        let layout = DescriptorSetLayout::builder()
            .add_layout_binding(
                0,
                VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER,
                VK_SHADER_STAGE_CLOSEST_HIT_BIT_KHR | VK_SHADER_STAGE_COMPUTE_BIT,
                0,
            )
            .build(device.clone())?;

        let descriptor_pool = DescriptorPool::builder()
            .set_layout(layout)
            .build(device.clone())?;

        let descriptor_set = DescriptorPool::prepare_set(&descriptor_pool).allocate()?;

        let info_buffer = Buffer::builder()
            .set_usage(VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT)
            .set_memory_usage(MemoryUsage::CpuOnly)
            .set_data(&[LightInfo { light_count: 0 }])
            .build(device.clone())?;

        descriptor_set.update(&[DescriptorWrite::uniform_buffers(0, &[&info_buffer])])?;

        Ok((info_buffer, descriptor_set))
    }

    fn present_image(
        &self,
        buffer_recorder: &mut CommandBufferRecorder<'_>,
        current_swapchain_image: &Arc<Image>,
    ) -> Result<()> {
        // bring image into transfer source mode
        buffer_recorder.image_barrier(
            &self.target_image,
            VK_IMAGE_LAYOUT_GENERAL,
            VK_PIPELINE_STAGE_ALL_COMMANDS_BIT,
            VK_IMAGE_LAYOUT_TRANSFER_SRC_OPTIMAL,
            VK_PIPELINE_STAGE_TRANSFER_BIT,
        );

        // transition swapchain image to transfer destination
        buffer_recorder.image_barrier(
            current_swapchain_image,
            VK_IMAGE_LAYOUT_PRESENT_SRC_KHR,
            VK_PIPELINE_STAGE_ALL_COMMANDS_BIT,
            VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL,
            VK_PIPELINE_STAGE_TRANSFER_BIT,
        );

        buffer_recorder.blit_complete(
            &self.target_image,
            current_swapchain_image,
            VK_FILTER_LINEAR,
        );

        // bring image into receive mode
        buffer_recorder.image_barrier(
            &self.target_image,
            VK_IMAGE_LAYOUT_TRANSFER_SRC_OPTIMAL,
            VK_PIPELINE_STAGE_TRANSFER_BIT,
            VK_IMAGE_LAYOUT_GENERAL,
            VK_PIPELINE_STAGE_ALL_COMMANDS_BIT,
        );

        // transition swapchain image to present mode
        buffer_recorder.image_barrier(
            current_swapchain_image,
            VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL,
            VK_PIPELINE_STAGE_TRANSFER_BIT,
            VK_IMAGE_LAYOUT_PRESENT_SRC_KHR,
            VK_PIPELINE_STAGE_ALL_COMMANDS_BIT,
        );

        Ok(())
    }

    #[inline]
    fn update_lights(&mut self) -> Result<()> {
        let buffers: Vec<&Arc<Buffer<LightData>>> =
            self.lights.iter().map(|l| l.buffer()).collect();

        self.light_descriptor_set
            .update(&[DescriptorWrite::storage_buffers(0, &buffers)])?;

        let mut mapped = self.light_info_buffer.map_complete()?;
        mapped[0].light_count = self.lights.len() as u32;

        Ok(())
    }

    fn gather_descriptor_buffers(
        content: &[&EntityObject],
    ) -> Result<(Vec<Arc<Image>>, Vec<Arc<Buffer<Material>>>, VkDeviceSize)> {
        let mut texture_array = Vec::new();
        let mut material_array = Vec::new();

        let mut collector = |color_texture: &Option<Arc<Image>>,
                             normal_texture: &Option<Arc<Image>>,
                             material_buffer: &Arc<Buffer<Material>>|
         -> Result<()> {
            let mut mapped = material_buffer.map_complete()?;

            mapped[0].color_texture_index = match color_texture {
                Some(ct) => {
                    let index = texture_array.len() as i32;
                    texture_array.push(ct.clone());

                    index
                }
                None => -1,
            };

            mapped[0].normal_texture_index = match normal_texture {
                Some(nt) => {
                    let index = texture_array.len() as i32;
                    texture_array.push(nt.clone());

                    index
                }
                None => -1,
            };

            material_array.push(material_buffer.clone());

            Ok(())
        };

        let mut vertex_count = 0;

        for entity in content.iter() {
            if let Ok(draw) = entity.get_component::<Draw>() {
                Self::collect_buffers(draw, &mut collector, &mut vertex_count)?;
            }
        }

        Ok((texture_array, material_array, vertex_count))
    }

    fn collect_buffers<F>(
        draw: &Draw,
        collector_function: &mut F,
        vertex_count: &mut u64,
    ) -> Result<()>
    where
        F: FnMut(&Option<Arc<Image>>, &Option<Arc<Image>>, &Arc<Buffer<Material>>) -> Result<()>,
    {
        for mesh in draw.iter() {
            for primitive_type in &mesh.primitive_types {
                match &primitive_type {
                    PrimitiveDataTypes::PositionOnly(primitive) => {
                        Self::collect_primitive(collector_function, vertex_count, primitive)?;
                    }
                    PrimitiveDataTypes::PositionNormal(primitive) => {
                        Self::collect_primitive(collector_function, vertex_count, primitive)?;
                    }
                    PrimitiveDataTypes::PositionNormalColor(primitive) => {
                        Self::collect_primitive(collector_function, vertex_count, primitive)?;
                    }
                    PrimitiveDataTypes::PositionNormalBone(primitive) => {
                        Self::collect_primitive(collector_function, vertex_count, primitive)?;
                    }
                    PrimitiveDataTypes::PositionColorUV(primitive) => {
                        Self::collect_primitive(collector_function, vertex_count, primitive)?;
                    }
                    PrimitiveDataTypes::PositionNormalColorUV(primitive) => {
                        Self::collect_primitive(collector_function, vertex_count, primitive)?;
                    }
                    PrimitiveDataTypes::PositionNormalColorUVNormalUV(primitive) => {
                        Self::collect_primitive(collector_function, vertex_count, primitive)?;
                    }
                    PrimitiveDataTypes::PositionNormalColorUVBone(primitive) => {
                        Self::collect_primitive(collector_function, vertex_count, primitive)?;
                    }
                    PrimitiveDataTypes::PositionNormalColorUVNormalUVBone(primitive) => {
                        Self::collect_primitive(collector_function, vertex_count, primitive)?;
                    }
                }
            }
        }

        Ok(())
    }

    fn collect_primitive<F, P: ReprC>(
        collector_function: &mut F,
        vertex_count: &mut u64,
        primitive: &PrimitiveData<P>,
    ) -> Result<()>
    where
        F: FnMut(&Option<Arc<Image>>, &Option<Arc<Image>>, &Arc<Buffer<Material>>) -> Result<()>,
        P: Copy,
    {
        *vertex_count += primitive.vertex_buffer.size();

        collector_function(
            &primitive.color_texture,
            &primitive.normal_texture,
            &primitive.material_buffer,
        )?;

        Ok(())
    }
}

impl<T> ExtensionCheck for RayTracer<T>
where
    T: RenderingCore,
{
    fn check(device: &Arc<Device>) -> bool {
        let physical_device = device.physical_device();

        if physical_device
            .descriptor_indexing_features()
            .descriptorBindingVariableDescriptorCount
            == VK_FALSE
        {
            return false;
        }

        T::check(device)
    }
}

impl<T> RenderingFrontEnd for RayTracer<T>
where
    T: RenderingCore,
{
    fn render_scale(&self) -> f32 {
        self.render_scale
    }

    fn view_mut(&mut self) -> &mut View {
        &mut self.view
    }

    fn view(&self) -> &View {
        &self.view
    }

    fn screen_to_world(&self, x: u32, y: u32) -> Result<Option<Vector3<f32>>> {
        let position = self.position_buffer.read_pixel(x, y)?;

        Ok(
            if !position.x.is_finite() || !position.y.is_finite() || !position.z.is_finite() {
                None
            } else {
                Some(position)
            },
        )
    }

    fn set_center(&mut self, center: Vector3<f32>) -> Result<()> {
        if self.view.camera_mut().set_center(center) {
            self.view.update_buffer()?;
        }

        Ok(())
    }

    fn resize(
        &mut self,
        window_width: f32,
        window_height: f32,
        _images: &TargetMode<Vec<Arc<Image>>>,
    ) -> Result<()> {
        let fov = 45.0;
        let proj = perspective(fov, window_width / window_height, 0.01, 100.0);
        self.view = View::new_proj(&self.view, proj, fov)?;

        let render_scale = self.render_scale;

        let width = (window_width * render_scale) as u32;
        let height = (window_height * render_scale) as u32;

        self.width = width;
        self.height = height;

        self.target_image = Self::create_target_image(&self.device, &self.queue, width, height)?;
        self.position_image =
            Self::create_position_image(&self.device, &self.queue, width, height)?;

        self.renderer
            .resize(&self.target_image, &self.position_image)?;

        Ok(())
    }

    // scene rendering
    fn process(
        &mut self,
        content: Vec<&EntityObject>,
        buffer_recorder: &mut CommandBufferRecorder<'_>,
        images: &TargetMode<Vec<Arc<Image>>>,
        indices: &TargetMode<usize>,
    ) -> Result<()> {
        if content.is_empty() {
            return Ok(());
        }

        // invalidate all command_buffers
        self.renderer.invalidate();
        self.animator.invalidate();

        // gather textures and materials from all objects
        let (textures, materials, vertex_count) = Self::gather_descriptor_buffers(&content)?;

        let mut recompile = false;

        // update material descriptor
        if !materials.is_empty() {
            let materials_ref: Vec<&Arc<Buffer<Material>>> = materials.iter().collect();

            if materials_ref.len() as u32 > self.max_material_count {
                while materials_ref.len() as u32 > self.max_material_count {
                    self.max_material_count *= 2;
                }

                self.material_desc_set =
                    Self::create_material_descriptor_set(&self.device, self.max_material_count)?;

                recompile = true;
            }

            self.material_desc_set
                .update(&[DescriptorWrite::uniform_buffers(0, &materials_ref)])?;
        }

        // update texture descriptor
        if !textures.is_empty() {
            let textures_ref: Vec<&Arc<Image>> = textures.iter().collect();

            if textures_ref.len() as u32 > self.max_texture_count {
                while textures_ref.len() as u32 > self.max_texture_count {
                    self.max_texture_count *= 2;
                }

                self.texture_desc_set =
                    Self::create_texture_descriptor_set(&self.device, self.max_texture_count)?;

                recompile = true;
            }

            self.texture_desc_set
                .update(&[DescriptorWrite::combined_samplers(0, &textures_ref)])?;
        }

        if recompile {
            self.renderer.recompile(&[
                &self.scene_desc_set,
                &self.material_desc_set,
                &self.texture_desc_set,
                &self.light_descriptor_set,
                &self.light_info_desc_set,
            ])?;
        }

        let index = *indices.mono();

        if let Some(scene_buffer) =
            self.animator
                .animate(&content, vertex_count, buffer_recorder, index, &self.view)?
        {
            self.scene_desc_set
                .update(&[DescriptorWrite::storage_buffers(0, &[scene_buffer])])?;
        }

        self.renderer.process(
            buffer_recorder,
            &[
                &self.scene_desc_set,
                &self.material_desc_set,
                &self.texture_desc_set,
                &self.light_descriptor_set,
                &self.light_info_desc_set,
            ],
            self.animator.buffer(),
            index,
        )?;

        self.present_image(buffer_recorder, &images.mono()[index])?;

        self.position_buffer.update(&self.position_image)?;

        Ok(())
    }

    // light handling
    fn add_light(&mut self, light: Light) -> Result<()> {
        debug_assert!(!self.lights.iter().any(|l| l.eq(&light)));

        self.lights.push(RayTracerLight::new(&self.device, light)?);
        self.update_lights()?;

        Ok(())
    }

    fn remove_light(&mut self, light: Light) -> Result<()> {
        if let Some(i) = self.lights.iter().position(|pl| pl.eq(&light)) {
            self.lights.remove(i);

            if !self.lights.is_empty() {
                self.update_lights()?;
            }
        }

        Ok(())
    }

    fn clear_lights(&mut self) -> Result<()> {
        self.lights.clear();

        Ok(())
    }
}
